/* --------------------------------------------------------------------
 predsmomtl_mex.c: MEX-file for SMOMTL prediction using model from GSMO.

 Compile:  mex -largeArrayDims predsmomtl_mex.c

 Synopsis:

  [pred_label, pred_error] = predsmomtl_mex(test,ker,arg,gamma,C,Alpha,groupid,d,svX,svy,nsv)

 compute the predicition error and the output for testing data using model
 from Generalized Sequential Minimal Optimization

 by Feng Cai, Oct. 16, 2009
 Prof. Vladimir Cherkassky's machine learning group at the University of 
 Minnesota, Twin Cities.
 
 Inputs:
   test
     test(1), test(2), ... test(t): binary labeled testing vectors for 
	 different groups.
   test(r) [struct] Binary labeled testing vectors for group r:
     .X [num_data x dim] Training vectors.
     .y [num_data x 1] Labels (1 or -1).

   
   ker [1x2] Kernel identifier for two spaces (default 0);
	 ker[0] kernel identifier for decision space
     ker[1] kernel identifier for correction space
       0 -- linear: u'*v
       1 -- radial basis function: exp(-|u-v|^2/(2*sigma^2))
   arg [1x2] Kernel argument(s) (default 1).
     arg[0] kernel argument for decision space
     arg[1] kernel argument for correction space
   gamma   regularization parameter 
   C       Regularization constant (default C=inf). 
   Alpha   alpha for support vectors
   groupid groupd id for support vectors
   d       bias terms for different groups
   svX     support vectors
   nsv     number of support vectors
   

 Output:
   pred_label predicted labels for different groups
   pred_error prediction error
*/

#include "mex.h"
#include "matrix.h"
#include <math.h>
#include <stdlib.h>
#include <string.h>

#define MAX(A,B)   (((A) > (B)) ? (A) : (B) )
#define MIN(A,B)   (((A) < (B)) ? (A) : (B) )


// --- Global variables ---------------------------------------------- 

unsigned long N = 0;       // number of testing patterns 
long DIM;                  // dimension of each instance 
double C;				   // trade-off constants 
double GAMMA;              // Regularization parameter
double **DATA;              // pointer at testing instances 
double *TARGET;            // pointer at labels of testing examples 
long *GROUPID;              // group id for each testing instance 
long *groupid;              // group id for support vectors
long nsv;                  // no. of support vectors
double **svX;               // support vectors
double *svy;               // labels of support vectors
double *ALPHA;             // Lagrange multipliers 
double *D;                 // Bias (threshold) for each group 
int KER[2];                // KER[0] kernel for decision space
						   //      0: linear kernel
						   //   	 1: rbf kernel
						   // KER[1] kernel for correction space
						   //     0: linear kernel
						   //    1: rbf kernel 
double ARG[2];             // ARG[0] parameter of kernel for decision space
						   //   ARG[1] parameter of kernel for correction space 
double ERROR;              // Prediction error
int *S;                    // predicted labels


// ==============================================================
// Implementation of Generalized Sequential Minimal Optimizer (GSMO)
// ============================================================ 

// --------------------------------------------------------------
// Kernel funtion for mtl
// -------------------------------------------------------------- 
double mtlkernel(long i1, long i2, int ker, double arg)
{
	double c = 0.;
	long i;
	if( ker==0 )
	{
		// linear kernel
		for( i=0; i<DIM; i++ )
		{
			c = c + DATA[i1][i]*svX[i2][i];
		}
	}
	else
	{
		// rbf kernel
		for( i=0; i<DIM; i++ )
		{
			c = c + (DATA[i1][i]-svX[i2][i])*(DATA[i1][i]-svX[i2][i]);
		}
		c = exp(-0.5*c/(arg*arg));
	}
	return c;
}



// --------------------------------------------------------------
// Main predSMOMTL.
// ------------------------------------------------------------ 
void runPREDSMOMTL( void )
{
	long i,j;
	double s;
	ERROR = 0;
	for( i=0; i<N; i++ )
	{
		s = 0;
		for( j=0; j<nsv; j++ )
		{
			s = s + ALPHA[j]*svy[j]*mtlkernel(i, j, KER[0], ARG[0]);
			if( GROUPID[i]==groupid[j] )
			{
				s = s + 1/GAMMA*ALPHA[j]*svy[j]*mtlkernel(i, j, KER[1], ARG[1]);
			}
		}
		s = s + D[GROUPID[i]-1];
		if( (s>0 && TARGET[i]<0) || (s<0 && TARGET[i]>0) )
		{
			ERROR = ERROR + 1;
		}
		if( s>0 )
		{
			S[i] = 1;
		}
		else
		{
			S[i] = -1;
		}
	}
	ERROR = ERROR/N;
}



// ==============================================================
// Main MEX function - interface to Matlab.
// ============================================================ 

void mexFunction( int nlhs, mxArray *plhs[],
		  int nrhs, const mxArray*prhs[] )
{
	long i,j,k,t;
	long i1, i2;
	long r;
	double C1, C2;
	mwSize NStructElems;
	mwSize     ndim;
	mxArray    *tmp, *tmp2;
	double    *fout;
	double *arraytmp, *arraytmp2, *arraytmp3;
	N = 0;

	//---- get input arguments ------------------------
	if( nrhs!=11 )
	{
		mexErrMsgTxt("11 inputs required.");
	}
	//else if( nlhs!=2 )
	//{
	//	mexErrMsgTxt("2 outputs required.");
	//}

	if( !mxIsStruct(prhs[0]) )
	{
		mexErrMsgTxt("Input must be a structure.");
	}

    NStructElems = mxGetNumberOfElements(prhs[0]);

	for( i=0; i<NStructElems; i++ )
	{
		tmp = mxGetField(prhs[0], i, "X");
		DIM = mxGetN(tmp);
		j = mxGetM(tmp);
		N = N + j;
	}
	
	// allocate memory for DATA
	if( (DATA = mxCalloc(N, sizeof(*DATA) )) == NULL) 
	{
		mexErrMsgTxt("Not enough memory.");
	}
	for( i=0; i<N; i++ )
	{
		if( (DATA[i] = mxCalloc(DIM, sizeof(double) )) == NULL) 
		{
			mexErrMsgTxt("Not enough memory.");
		}
	}

	// allocate memory for S
	if( (S = mxCalloc(N, sizeof(int) )) == NULL) 
	{
		mexErrMsgTxt("Not enough memory.");
	}

	// allocate memory for TARGET (labels) (1,-1) 
	if( (TARGET = mxCalloc(N, sizeof(double) )) == NULL) 
	{
		mexErrMsgTxt("Not enough memory.");
	}

	// allocate memory for GROUPID
	if( (GROUPID = mxCalloc(N, sizeof(int) )) == NULL) 
	{
		mexErrMsgTxt("Not enough memory.");
	}

	k = 0;
	for( i=0; i<NStructElems; i++ )
	{
		tmp = mxGetField(prhs[0], i, "X");
		DIM = mxGetN(tmp);
		j = mxGetM(tmp);
		arraytmp = mxGetPr(tmp);
		tmp2 = mxGetField(prhs[0], i, "y");
		arraytmp2 = mxGetPr(tmp2);
		
		for( i1=0; i1<j; i1++ )
		{
			for( i2=0; i2<DIM; i2++ )
			{
				DATA[i1+k][i2] = arraytmp[i2*j+i1];
			}

			GROUPID[i1+k] = i+1;
			TARGET[i1+k] = arraytmp2[i1];
		}
		k = k + j;
	}

	// get value for KER
	arraytmp = mxGetPr( prhs[1] );
	KER[0] = arraytmp[0];
	KER[1] = arraytmp[1];

	// get value for ARG
	arraytmp = mxGetPr( prhs[2] );
	ARG[0] = arraytmp[0];
	ARG[1] = arraytmp[1];

	// get value for GAMMA
	arraytmp = mxGetPr( prhs[3] );
	GAMMA = *arraytmp;

	// get value for C
	arraytmp = mxGetPr( prhs[4] );
	C = *arraytmp;

	// get value for ALPHA
	ALPHA = mxGetPr( prhs[5] );

	// get value for nsv
	arraytmp = mxGetPr( prhs[10] );
	nsv = (long)*arraytmp;

	// allocate memory for groupid
	if( (groupid = mxCalloc(nsv, sizeof(int) )) == NULL) 
	{
		mexErrMsgTxt("Not enough memory.");
	}
	// get value for groupid
	arraytmp = mxGetPr( prhs[6] );
	for( i=0; i<nsv; i++ )
	{
		groupid[i] = (long)arraytmp[i];
	}

	// get value for D
	D = mxGetPr( prhs[7] );

	// get value for svy
	svy = mxGetPr( prhs[9] );


	// get value for svX
	arraytmp = mxGetPr( prhs[8] );
	DIM = mxGetN( prhs[8] );

	// allocate memory for svX
	if( (svX = mxCalloc(nsv, sizeof(*svX) )) == NULL) 
	{
		mexErrMsgTxt("Not enough memory.");
	}
	for( i=0; i<nsv; i++ )
	{
		if( (svX[i] = mxCalloc(DIM, sizeof(double) )) == NULL) 
		{
			mexErrMsgTxt("Not enough memory.");
		}
	}
	for( i=0; i<nsv; i++ )
	{
		for( j=0; j<DIM; j++ )
		{
			svX[i][j] = arraytmp[j*nsv+i];
		}
	}


	// ---------- run PREDSMOMTL --------------------
	runPREDSMOMTL();

	// ---------- outputs --------------------------
	plhs[0] = mxCreateDoubleMatrix(N,2,mxREAL);
	arraytmp = mxGetPr( plhs[0] );
	for( i=0; i<N; i++ )
	{
		arraytmp[i] = S[i];
		arraytmp[i+N] = GROUPID[i];
	}

	plhs[1] = mxCreateDoubleMatrix(1,1,mxREAL);
	arraytmp = mxGetPr( plhs[1] );
	*arraytmp = ERROR;

	for( i=0; i<nsv; i++)
	{
		mxFree( svX[i] );
	}
	mxFree( (void *)svX );

	for( i=0; i<N; i++)
	{
		mxFree( DATA[i] );
	}
	mxFree( (void *)DATA );
	mxFree( TARGET );
	mxFree( GROUPID );
	mxFree( groupid );
	mxFree( S );
}